﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.PDPSimulation.DomainModel;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.PDPSimulation {
  [Item("PriorityDispatching", "")]
  [StorableClass]
  public sealed class PriorityDispatching : DynamicPDPOptimization {

    private bool IsIdle(Vehicle v) {
      return v.CurrentOrder == Guid.Empty;
    }

    protected override bool OptimizationRequired(ChangeInformation changeInformation) {
      // either there are vehicles waiting and there are orders to dispatch
      // or there are new orders and waiting vehicles
      return (GetVehicles().Any(x => IsIdle(x))
        && GetOrders().Any(x => x.AssignedVehicle == Guid.Empty || GetVehicles().Any(y => IsIdle(y) && x.AssignedVehicle == y.Id)));
    }

    [StorableConstructor]
    private PriorityDispatching(bool deserializing) : base(deserializing) { }
    private PriorityDispatching(PriorityDispatching original, Cloner cloner) : base(original, cloner) { }
    public PriorityDispatching() {

    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new PriorityDispatching(this, cloner);
    }

    protected override bool PerformOptimization(DynamicPDProblemInstance instance, ChangeInformation changeInformation) {
      var waitingVehicles = GetVehicles().Where(x => IsIdle(x)).ToList();
      var waitingOrders = GetOrders().Where(x => x.AssignedVehicle == Guid.Empty).ToList();

      while (waitingVehicles.Any() && (waitingOrders.Any()
        || GetOrders().Any(x => x.OrderState == OrderState.PickedUp && waitingVehicles.Any(y => y.Id == x.AssignedVehicle)))) {
        var tmp = new Dictionary<Vehicle, Tuple<Order, double>>();
        foreach (var vehicle in waitingVehicles) {
          var orders = waitingOrders.Concat(GetOrders().Where(x => x.OrderState == OrderState.PickedUp && x.AssignedVehicle == vehicle.Id));
          if (orders.Any()) {
            Order selectedOrder;
            double priority;
            GetHighestPriorityOrder(vehicle, orders, out selectedOrder, out priority);
            tmp[vehicle] = Tuple.Create(selectedOrder, priority);
          }
        }
        var assignment = tmp.GroupBy(x => x.Value.Item1, x => new { Vehicle = x.Key, Priority = x.Value.Item2 });
        foreach (var a in assignment) {
          var bestVehicle = a.MaxItems(x => x.Priority).Select(x => x.Vehicle).First();
          var order = a.Key;

          MoveVehicleAction action;
          if (order.OrderState == OrderState.Waiting) {
            action = new MoveVehicleAction(bestVehicle.Id, order.PickupXCoord, order.PickupYCoord, VehicleAction.Pickup, order, scenario.AllowDiversion, scenario.DistanceMeasure);
          } else {
            action = new MoveVehicleAction(bestVehicle.Id, order.DeliveryXCoord, order.DeliveryYCoord, VehicleAction.Deliver, order, scenario.AllowDiversion, scenario.DistanceMeasure);
          }
          SetAction(action);
          waitingVehicles.Remove(bestVehicle);
          waitingOrders.Remove(order);

          if (GetOrders().Where(o => o.AssignedVehicle == bestVehicle.Id && o.OrderState == OrderState.Waiting).Any()) {
            throw new InvalidOperationException("An order has already been assigned to the vehicle");
            //can be removed after debugging
          }

          order.AssignedVehicle = bestVehicle.Id;
        }
      }
      return true;
    }

    private void GetHighestPriorityOrder(Vehicle vehicle, IEnumerable<Order> orders, out Order order, out double priority) {
      var random = new System.Random();
      var o = orders.ToArray();
      order = o[random.Next(o.Length)];
      priority = random.NextDouble();
    }

    /*private void GetHighestPriorityOrder(Vehicle vehicle, IEnumerable<Order> orders, out Order order, out double priority) {
      var pdp = instance as DynPDPProblemInstance;
      if (pdp == null) throw new ArgumentException("instance is not of type DynPDPProblemInstance", "instance");

      int vehicles = pdp.Vehicles.Value;
      while (solution.Unrouted.Count > 0) {
        var diversion = new Dictionary<int, Dictionary<int, double>>(vehicles);
        foreach (var v in Enumerable.Range(0, vehicles)) diversion[v] = new Dictionary<int, double>();
        var estimatedTimeOfArrival = new Dictionary<int, Dictionary<int, double>>(vehicles);
        var earliestDueDate = new Dictionary<int, double>(solution.Unrouted.Count);
        var earliestStartTime = new Dictionary<int, double>(solution.Unrouted.Count);
        var numberOfWaitingOrders = new Dictionary<int, double>(solution.Unrouted.Count);
        var                                                                                                      
        foreach (var s in solution.Unrouted) {
          
        }
      }


      double FUTURE = 1e07;
      int locations = myEnvironment.servicePoints.length;
      // these are the dispatching rules that are calculated for each location to decide on the next destination
      double[] distance = new double[locations]; // distance from current position to location
      double[] estimatedTimeOfArrival = new double[locations];
      double[] earliestDueDate = new double[locations]; // difference to current time that a passenger needs to arrive which is currently carried
      double[] earliestStartTime = new double[locations]; // difference to current time when a waiting passenger requested pickup
      int[] numberOfWaitingOrders = new int[locations];
      int[] numberOfWaitingOrderItems = new int[locations];
      int[] numberOfDueOrders = new int[locations];
      int[] numberOfDueOrderItems = new int[locations];
      double[] averageDistanceToOtherDestinations = new double[locations]; // The average distance of a location to all the destinations that have to be visited
      double[] minDistanceToOtherDestinations = new double[locations]; // The min distance of a location to all the destinations that have to be visited
      double[] maxDistanceToOtherDestinations = new double[locations]; // The max distance of a location to all the destinations that have to be visited
      int[] numberOfCouriersWithThisDestination = new int[locations]; // How many other vehicles already go there
      double[] averageDistanceToOtherCouriers = new double[locations];
      double[] minDistanceToOtherCouriers = new double[locations];
      double[] maxDistanceToOtherCouriers = new double[locations];

      // calculate each dispatching rule
      for (int i = 0; i < locations; i++) {
        if (i != currentDestination) {
          ServicePoint servicePoint = myEnvironment.servicePoints[i];

          distance[i] = myEnvironment.distances[i][currentDestination];
          estimatedTimeOfArrival[i] = distance[i] / myEnvironment.averageSpeed;

          double earliestDue = FUTURE;
          double minDist = 1e07, maxDist = 0, distSum = 0.0;
          int ordersDue = 0, itemsDue = 0;
          for (int x = 0; x < currentOrders.size(); x++) {
            if (currentOrders.get(x).dropOffAtId == i) {
              ordersDue++;
              itemsDue += currentOrders.get(x).demandSize;
              if ((currentOrders.get(x).dueDate - now) < earliestDue) earliestDue = currentOrders.get(x).dueDate - now;
            }
            double d = myEnvironment.distances[i][currentOrders.get(x).dropOffAtId];
            distSum += d;
            if (d < minDist) minDist = d;
            if (d > maxDist) maxDist = d;
          }
          earliestDueDate[i] = earliestDue;
          numberOfDueOrders[i] = ordersDue;
          numberOfDueOrderItems[i] = itemsDue;
          minDistanceToOtherDestinations[i] = ((currentOrders.size() == 0) ? (0) : (minDist));
          maxDistanceToOtherDestinations[i] = ((currentOrders.size() == 0) ? (0) : (maxDist));
          averageDistanceToOtherDestinations[i] = ((currentOrders.size() == 0) ? (0) : (distSum / (double)currentOrders.size()));

          numberOfWaitingOrders[i] = servicePoint.orders.size();
          double earliestStart = FUTURE;
          int items = 0;
          for (int k = 0; k < numberOfWaitingOrders[i]; k++) {
            if ((servicePoint.orders.get(k).startTime - now) < earliestStart) earliestStart = servicePoint.orders.get(k).startTime - now;
            items += servicePoint.orders.get(k).demandSize;
          }
          numberOfWaitingOrderItems[i] = items;
          earliestStartTime[i] = earliestStart;
          numberOfCouriersWithThisDestination[i] = 0;
          minDist = 1e07; maxDist = 0; distSum = 0.0;
          for (int m = 0; m < myEnvironment.agents.size(); m++) {
            CourierAgent agent = myEnvironment.agents.get(m);
            if (agent != this) {
              if (agent.nextDestination == i) numberOfCouriersWithThisDestination[i]++;
              double d = myEnvironment.distances[i][agent.nextDestination];
              distSum += d;
              if (minDist > d) minDist = d;
              if (maxDist < d) maxDist = d;
            }
          }
          minDistanceToOtherCouriers[i] = minDist;
          maxDistanceToOtherCouriers[i] = maxDist;
          averageDistanceToOtherCouriers[i] = distSum / (double)myEnvironment.agents.size();
        }
      }

      double[] priorityList = new double[locations];
      for (int i = 0; i < locations; i++) {
        if (i != currentDestination) {
          priorityList[i] = DIST * distance[i]
                          + ETA * estimatedTimeOfArrival[i]
                  + EDD * earliestDueDate[i]
                  + EST * earliestStartTime[i]
                  + NWO * numberOfWaitingOrders[i]
                  + NDO * numberOfDueOrders[i]
                  + NWOI * numberOfWaitingOrderItems[i]
                  + NDOI * numberOfDueOrderItems[i]
                  + ADD * averageDistanceToOtherDestinations[i]
                  + MiDD * minDistanceToOtherDestinations[i]
                  + MaDD * maxDistanceToOtherDestinations[i]
                  + NCD * numberOfCouriersWithThisDestination[i]
                  + ADC * averageDistanceToOtherCouriers[i]
                  + MiDC * minDistanceToOtherCouriers[i]
                  + MaDC * maxDistanceToOtherCouriers[i];
        } else priorityList[i] = Double.MAX_VALUE;
      }
      double minPrio = priorityList[0];
      int minPrioIdx = 0;
      //System.out.print(minPrio);
      for (int i = 1; i < locations; i++) {
        //System.out.print(", " + priorityList[i]);
        if (priorityList[i] < minPrio) {
          minPrio = priorityList[i];
          minPrioIdx = i;
        }
      }
      //System.out.println();
      //System.out.println("Picked " + minPrioIdx);

      nextDestination = minPrioIdx;
    }*/
  }
}
